! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: embadpixfind_mod
!
! Authors:  Jean Ballet - SAp   CEA Saclay
!          
! Description: Find EPIC-MOS bad pixels
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module embadpixfind_mod

  use emutils_mod
  use badpixutils_mod
  use cal
  use dal
  use param  
  use errorHandling

  implicit none 
  
  private 

! Names of the task parameters (2nd string for the keywords)

    character(len=*), parameter :: outimage    = "evimageset", &
                                   outbadpix   = "badpixset", &
                                   incbadpix   = "incremental", &
                                   threshold   = "probathreshold", &
                                   badcols     = "findbadsegments", &
                                   onbright    = "findbright", &
                                   ondead      = "finddead", &
                                   hwidth2D    = "halfwidth2d", &
                                   hwidth1D    = "halfwidth1d", &
                                   maxrat      = "maxratio",  &
                                   minrat      = "minratio",  &
                                   iterate     = "niter",  &
                                   usecal      = "usecal", &
                                   nobrightccf = "ignoreccfbright", &
                                   adddead     = "includedeadpixels"

! Maximum number of bad pixels per CCD

    integer,parameter :: maxbad = 32768

! Keywords to check for compatibility of image file and bad pixels file

    character(len=8), dimension(4), parameter :: keycompat = &
         (/"TELESCOP","INSTRUME","CCDID   ","CCDNODE "/)

! Window parameters

    integer :: wx0, wx1, wy0, wy1

  public :: embadpixfind

contains 
  
  !---------------------------------------------------------------------
  subroutine embadpixfind
  !---------------------------------------------------------------------

    type(ArrayT) :: ev_map
    type(TableT) :: bad_tab
    integer(kind=int16), dimension(maxbad) :: xbad, ybad, yext, tbad, fbad
    integer :: nbad, niter, w2, w21D, node
    real    :: maxproba, maxratio, minratio
    logical :: findcols, finddead, findbright, incremental
    logical :: withcal, withdead, withbrightccf

! Read parameters

    call getParamValues(ev_map, bad_tab, incremental, maxproba, &
                        maxratio, minratio, findcols, findbright, finddead,  &
                        w2, w21D, niter, withcal, withdead, withbrightccf)

    call readBadpix(bad_tab, incremental, xbad, ybad, yext, tbad, fbad, nbad)

! New bad pixels are not from the CCF

    fbad(nbad+1:maxbad) = 3

! Look for bad pixels

    call findAllBad(ev_map, maxproba, maxratio, minratio, findcols,  &
                    findbright, finddead, niter, withcal, withdead,  &
                    withbrightccf, w2, w21D, xbad, ybad, yext, tbad, fbad,nbad)

! Merge bad pixels aligned along a column

    if (nbad > 1)  call mergeBad(xbad, ybad, yext, tbad, fbad, nbad)

! Write bad pixels file
    if ( (trim(stringAttribute(bad_tab,"INSTRUME")) .eq. 'RGS1') .or. &
        (trim( stringAttribute(bad_tab,"INSTRUME") ) .eq. 'RGS2') ) then
            node = int32Attribute(ev_map,"CCDNODE")
            if (node == 1) then
            xbad = (342 - xbad)+1
            call deleteAttribute(bad_tab,"WINDOWX0")
            call setAttribute(bad_tab,"WINDOWX0",172, comment="Lenght..")
        endif
    endif

    call writeBadpix(bad_tab, xbad, ybad, yext, tbad, fbad, nbad)

    call release(parent(bad_tab))
    call release(parent(ev_map))

  end subroutine embadpixfind

! Read data and call all specific bad structure finders

  subroutine findAllBad(ev_map, maxproba, maxratio, minratio,  &
                        findcols, findbright, finddead, niter, withcal,  &
                        withdead, withbrightccf, w2, w21D,  &
                        xbad, ybad, yext, tbad, fbad, nbad)

! ev_map     : Image array
! maxproba   : Probability threshold for false detection
! maxratio   : Relative detection threshold for dark features
! minratio   : Relative detection threshold for bright features
! findcols   : Look for bad rows or columns if true
! findbright : Look for bright pixels if true
! finddead   : Look for dead pixels if true
! niter      : Number of iterations
! withcal    : CAL_getBadPixelList is called if true
! withdead   : Dead pixels from the CAL are written to output list if true
! withbrightccf: Bright pixels in CCF are ignored in the analysis
! w2         : Half width of window for 2D search
! w21D       : Half width of window for 1D search
! xbad       : array of RAWX coordinates
! ybad       : array of RAWY coordinates
! yext       : array of RAWY extensions
! tbad       : array of bad pixel types
! fbad       : array of bad pixel status (uplinked, CCF or new)
! nbad       : number of bad pixels

    type(ArrayT), intent(in)    :: ev_map
    real,         intent(in)    :: maxproba, maxratio, minratio
    logical,      intent(in)    :: withcal, withdead, withbrightccf,  &
                                   findcols, findbright, finddead
    integer,      intent(in)    :: niter, w2, w21D
    integer(kind=int16), dimension(:), intent(inout) :: xbad, ybad, yext,  &
                                                        tbad, fbad
    integer,      intent(inout) :: nbad

    integer(kind=int32), dimension(:,:), pointer     :: map
    real,                dimension(:,:), allocatable :: medsmooth, badtest
    logical,             dimension(:,:), allocatable :: goodPixel, goodPixref
    integer :: nx, ny, ibad, iter, oldbad, nbad0, nbad1, width, wid1D
    integer :: x0, y0, y1, y2

! Define full widths

    width = 2*w2 + 1
    wid1D = 2*w21D + 1

! Read original and smoothed arrays

    map => int32Array2Data(ev_map)

    nx = size(map,1)
    ny = size(map,2)

! Initialise the goodPixel array

    allocate(goodPixel(nx,ny))
    goodPixel = .true.

! Set dead and uplinked pixels to 0

    if (withcal)  &
       call getCalBadpix(map, withdead, withbrightccf, goodPixel,  &
                         xbad,ybad,yext,tbad,fbad,nbad)

! return immediately if no search is asked for


    if (.not. (findbright .or. finddead)) then
       deallocate(goodPixel)
       return
    endif

! Do not consider already known bad pixels
! This must be after getCalBadpix for the check of uplinked bright pixels

    do ibad = 1, nbad
       x0 = xbad(ibad)
       y0 = ybad(ibad)
       goodPixel(x0,y0:y0+yext(ibad)-1) = .false.
    enddo

! Remember old bad pixels
    nbad0 = nbad

! Set all pixels outside window to bad

    if (wx0 >  1)  goodPixel( 1:wx0-1,:) = .false.
    if (wx1 < nx)  goodPixel(wx1+1:nx,:) = .false.
    if (wy0 >  1)  goodPixel(:, 1:wy0-1) = .false.
    if (wy1 < ny)  goodPixel(:,wy1+1:ny) = .false.

! Initialize secondary arrays

    allocate(medsmooth(nx,ny),badtest(nx,ny),goodPixRef(nx,ny))
    badtest = 0.
    call message(AppMsg,Verbose,"Applying median filter...")
    call avMedFilter(map,goodPixel,width,medsmooth,badtest,wx0,wx1,wy0,wy1)

! Iterate over the full loop several times

    do iter = 1,niter
       call message(AppMsg,Sparse,"")
       write(errstr,*) "Iteration", iter
       call message(AppMsg,Sparse)

! Look for bright pixels first to avoid disrupting bad columns/rows search

       goodPixRef = goodPixel
       oldbad = nbad
       call findBadPix(map, medsmooth, badtest, goodPixel, width, maxproba,  &
               maxratio, minratio, .false., xbad, ybad, yext, tbad, nbad)

! Do not include series of new bright pixels when looking for bright columns
! By construction yext is always 1 in output of findBadPix

       do ibad = oldbad+1, nbad
          x0 = xbad(ibad)
          y0 = ybad(ibad)
          y1 = max(y0-2,wy0)
          y2 = min(y0+2,wy1)
          if (count(.not. goodPixel(x0,y1:y2)) < 3)  &
             goodPixRef(x0,y0) = .false.
       enddo
       oldbad = nbad

! Look for bad columns and rows (both dark and bright)
! Always look for both types to avoid misleading the search

       if (findcols) then
          call findBadCol(map, goodPixRef, wid1D, maxproba, maxratio,  &
                          minratio, medsmooth, badtest, width,  &
                          xbad, ybad, yext, tbad, nbad)
          call findBadRow(map, goodPixRef, wid1D, maxproba, maxratio,  &
                          minratio, medsmooth, badtest, width,  &
                          xbad, ybad, yext, tbad, nbad)
       endif

! Go back to full mask

       goodPixel = goodPixel .and. goodPixRef

! Look for dark pixels

       if (finddead)  &
          call findBadPix(map,medsmooth,badtest,goodPixel,width,maxproba, &
                  maxratio, minratio, .true., xbad, ybad, yext, tbad, nbad)

! Stop if no new bad pixel is found in late stages

       if (nbad == oldbad)  exit
    enddo

! Remove very bright pixels and bright rows/columns from list if need be

    if (.not. findbright) then
       nbad1 = nbad0 + count(tbad(nbad0+1:nbad) == DEAD)
       xbad(nbad0+1:nbad1) = pack(xbad(nbad0+1:nbad),tbad(nbad0+1:nbad)==DEAD)
       ybad(nbad0+1:nbad1) = pack(ybad(nbad0+1:nbad),tbad(nbad0+1:nbad)==DEAD)
       yext(nbad0+1:nbad1) = pack(yext(nbad0+1:nbad),tbad(nbad0+1:nbad)==DEAD)
       tbad(nbad0+1:nbad1) = DEAD
       nbad = nbad1
    endif

! Remove dark rows/columns from list if need be

    if (.not. finddead) then
       nbad1 = nbad0 + count(tbad(nbad0+1:nbad) == HOT)
       xbad(nbad0+1:nbad1) = pack(xbad(nbad0+1:nbad),tbad(nbad0+1:nbad)==HOT)
       ybad(nbad0+1:nbad1) = pack(ybad(nbad0+1:nbad),tbad(nbad0+1:nbad)==HOT)
       yext(nbad0+1:nbad1) = pack(yext(nbad0+1:nbad),tbad(nbad0+1:nbad)==HOT)
       tbad(nbad0+1:nbad1) = HOT
       nbad = nbad1
    endif

    deallocate(goodPixel,goodPixRef,medsmooth,badtest)

  end subroutine findAllBad

! Find the bad pixels sticking out from the noise

  subroutine findBadPix(map, medsmooth, badtest, goodPixel, width, maxproba,  &
                maxratio, minratio, negative, xbad, ybad, yext, tbad, nbad)

! map       : Image array
! medsmooth : Same image after median smoothing
! badtest   : Significance map
! goodPixel : False for bad pixels
! width     : Width over which the comparison is done
! maxproba  : Probability threshold for false detection
! maxratio  : Relative detection threshold for dark features (< 1)
! minratio  : Relative detection threshold for bright features (> 1)
! negative  : Look for dead pixels instead if true
! xbad      : array of RAWX coordinates
! ybad      : array of RAWY coordinates
! yext      : array of RAWY extensions
! tbad      : array of bad pixel types
! nbad      : number of bad pixels

    integer(kind=int32), dimension(:,:), pointer       :: map
    real,                dimension(:,:), intent(inout) :: medsmooth, badtest
    logical,             dimension(:,:), intent(inout) :: goodPixel
    real,                                intent(in)    :: maxproba, maxratio, &
                                                          minratio
    integer,                             intent(in)    :: width
    logical,                             intent(in)    :: negative
    integer(kind=int16), dimension(:),   intent(inout) :: xbad,ybad,yext,tbad
    integer,                             intent(inout) :: nbad

    real,    external     :: erfc
    integer, dimension(2) :: maxpos
    real    :: proba, proba2, mu, mu2, minsigma, signif
    integer :: ncur, nbad0, w2
    integer :: ibad, i1, j1, i2, j2

    w2 = width / 2

! Get number of sigmas corresponding to maxproba

    minsigma = (quantilePoisson(1.e4,maxproba)-1.e4) / 100

    if (negative) then
       call message(AppMsg,Verbose,"Looking for dark pixels...")
    else
       call message(AppMsg,Verbose,"Looking for bright pixels...")
    endif

! No need to go further if all pixels are already set to bad

    if (.not. any(goodPixel))  return

    nbad0 = nbad
    do ibad = 1,size(xbad)-nbad

! Look for maximum local excess where goodPixel is True
! Should not be used with pointers
! mu and mu2 must be initialized to cover ncur < nwidth

       mu = 1
       if (negative) then
          maxpos = minloc(badtest,goodPixel)
          mu2 = 0
       else
          maxpos = maxloc(badtest,goodPixel)
          mu2 = 2 * minratio * mu
       endif
       signif = badtest(maxpos(1),maxpos(2))

! Stop if less than required significance

       if (abs(signif) < minsigma)  exit

       goodPixel(maxpos(1),maxpos(2)) = .false.
       write(errstr,*) "Investigating", maxpos, signif
       call message(AppMsg,Noisy)

! Define width x width reference zone

       i1 = max(maxpos(1)-w2,wx0)
       i2 = min(maxpos(1)+w2,wx1)
       j1 = max(maxpos(2)-w2,wy0)
       j2 = min(maxpos(2)+w2,wy1)
       ncur = count(goodPixel(i1:i2,j1:j2))
       if (ncur < width) then
          write(errstr,*) "Less than", width, " good pixels around", maxpos, &
                          ". Counted as bad."
          call message(AppMsg,Verbose)
       else

! Read median local value outside suspect pixel and value of suspect pixel
! erfc fails when its argument is larger than 9

          mu = medsmooth(maxpos(1),maxpos(2))
          mu2 = map(maxpos(1)-1,maxpos(2)-1)
          if (abs(signif) < 12.5) then
             proba = erfc(abs(signif)/sqrt(2.)) / 2
          else
             proba = 0.d0
          endif

! Compute exact significance using binomial law

          if (negative) then
             proba2 = cumulativeBinomial(nint(ncur*mu),nint(mu2),  &
                                         ncur/(ncur+1.d0))
          else
             proba2 = cumulativeBinomial(nint(mu2),nint(ncur*mu),1.d0/(ncur+1))
          endif

          proba = max(proba,proba2)
          write(errstr,*) "Pixel value:", mu2,  &
               "   Local average:", mu, "   Probability:", proba

          if (proba > maxproba) then

! Reinstate. Was not so bad after all.

             call message(AppMsg,Noisy)
             goodPixel(maxpos(1),maxpos(2)) = .true.
             badtest(maxpos(1),maxpos(2)) = 0.
             cycle
          else

! Recompute significance around removed pixel

             call message(AppMsg,Noisy)
             call avMedFilter(map, goodPixel, width, medsmooth, badtest,  &
                              i1, i2, j1, j2)
          endif
       endif

! Count dark pixels only if they are below maxratio * average
! Count bright pixels only if they are above minratio * average

       if (negative .and. mu2>maxratio*mu)  cycle
       if (.not. negative .and. mu2<minratio*mu) cycle
       nbad = nbad + 1
       xbad(nbad) = maxpos(1)
       ybad(nbad) = maxpos(2)
       yext(nbad) = 1
       if (negative) then
          tbad(nbad) = DEAD
       else
          tbad(nbad) = HOT
       endif
    enddo

    if (negative) then
       write(errstr,*) nbad-nbad0, " dark pixels found in image"
    else
       write(errstr,*) nbad-nbad0, " bright pixels found in image"
    endif
    call message(AppMsg,Sparse)

  end subroutine findBadPix

! Find bad columns and bad segments within
! Update medsmooth and badtest

  subroutine findBadCol(map, goodPixel, wid1D, maxproba, maxratio, minratio,  &
                medsmooth, badtest, width, xbad, ybad, yext, tbad, nbad)

! map       : Image array
! goodPixel : False for bad pixels
! wid1D     : Width over which the comparison is done on columns
! maxproba  : Probability threshold for false detection
! maxratio  : Relative detection threshold for dark features
! minratio  : Relative detection threshold for bright features
! medsmooth : Image after median smoothing
! badtest   : Significance map
! width     : Width over which the comparison is done on image
! xbad      : array of RAWX coordinates
! ybad      : array of RAWY coordinates
! yext      : array of RAWY extensions
! tbad      : array of bad pixel types
! nbad      : number of bad pixels

    integer(kind=int32), dimension(:,:), pointer       :: map
    logical,             dimension(:,:), intent(inout) :: goodPixel
    real,                dimension(:,:), intent(inout) :: medsmooth, badtest
    integer,                             intent(in)    :: wid1D, width
    real,                                intent(in)    :: maxproba, maxratio, &
                                                          minratio
    integer(kind=int16), dimension(:),   intent(inout) :: xbad,ybad,yext,tbad
    integer,                             intent(inout) :: nbad

    integer, dimension(:),   allocatable :: weight
    integer, dimension(1024)             :: cbad, ubad, xseg, lseg
    real,    dimension(:),   allocatable :: pixrate, profil, toCounts
    logical, dimension(:),   allocatable :: goodcol, goodcur
    integer :: nx, ny, nbc, nseg, nbd, ngood, w21D, w2
    integer :: ic, is, i, j ,i1, jc
    real    :: refcts, maxexp

    call message(AppMsg,Verbose,"Looking for bad columns...")

    w21D = wid1D / 2
    w2 = width / 2

! Project the image onto RAWX and look for bad columns
! Compensate for brightness profile along columns

    nx = size(map,1)
    ny = size(map,2)
    allocate(goodcur(wid1D))
    allocate(profil(nx),toCounts(nx),goodcol(nx))
    allocate(pixrate(ny),weight(ny))

    toCounts = 0.
    goodcol = any(goodPixel,2)
    do i = wx0, wx1
       if (.not. goodcol(i))  cycle
       i1 = min(max(i,wx0+w21D),wx1-w21D)
       weight = count(goodPixel(i1-w21D:i1+w21D,:),1)
       ngood = count(weight > 0)
       where (weight == 0)
          weight = 1
       endwhere
       pixrate = real(sum(map(i1-w21D-1:i1+w21D-1,:),1,  &
                          goodPixel(i1-w21D:i1+w21D,:)))
       refcts = sum(pixrate/weight) * (wy1-wy0+1) / ngood
       if (refcts > 1.e-10) then
          toCounts(i) = sum(pixrate/weight,goodPixel(i,:)) / refcts
       else
          toCounts(i) = -1.   ! Just a convention
       endif
    enddo

! Do not count entirely bad rows

    maxexp = maxval(toCounts)
    if (maxexp < 1)  toCounts = toCounts / maxexp
    toCounts = abs(toCounts)   ! set -1 back to 1 for empty columns

    profil = 0.
    do i = wx0, wx1
       if (toCounts(i) > 1.e-10)  &
          profil(i) = sum(map(i-1,:),goodPixel(i,:)) / toCounts(i)
    enddo

    call findBad1D(profil, toCounts, wid1D, maxproba, maxratio, minratio,  &
                   'column', cbad, ubad, nbc)

    nbd = count(ubad(1:nbc) == DEAD)
    write(errstr,*) nbd, " dark and", nbc-nbd, " bright columns found in image"
    call message(AppMsg,Sparse)
    deallocate(profil,toCounts,weight)

! Check whether only part of the column is bad

    pixrate = 0.
    do ic = 1, nbc
       write(errstr,*) "Column", cbad(ic)
       call message(AppMsg,Noisy)
       i1 = min(max(cbad(ic),wx0+w21D),wx1-w21D)

! Check number of nearby entirely bad columns

       if (count(goodcol(i1-w21D:i1+w21D)) < min(3,w21D)) then
          write(errstr,*) "next to other bad columns. Declared entirely bad."
          call message(AppMsg,Noisy)
          nbad = nbad + 1
          xbad(nbad) = cbad(ic)
          ybad(nbad) = wy0
          yext(nbad) = wy1 - wy0 + 1
          tbad(nbad) = ubad(ic)
          goodPixel(cbad(ic),wy0:wy1) = .false.
       else

! Get expected count rate for each pixel
! Do not use median because this is later summed over a range of pixels
! The median would give very poor results for small counts

          do j = wy0, wy1
             goodcur = goodPixel(i1-w21D:i1+w21D,j)

! Ignore nearby suspect columns starting at current one
! Previous ones should have been cleaned already

             do jc = ic, nbc
                if (abs(cbad(jc)-i1) <= w21D)  &
                   goodcur(w21D+1+cbad(jc)-i1) = .false.
             enddo
             pixrate(j) = real(sum(map(i1-w21D-1:i1+w21D-1,j-1),goodcur))  &
                        / max(count(goodcur),1)
          enddo

          call findBadSeg(map(cbad(ic)-1,:), goodPixel(cbad(ic),:),  &
                          pixrate, ubad(ic), xseg, lseg, nseg)
          do is = 1, nseg
             nbad = nbad + 1
             xbad(nbad) = cbad(ic)
             ybad(nbad) = xseg(is)
             yext(nbad) = lseg(is)
             tbad(nbad) = ubad(ic)
             goodPixel(cbad(ic),xseg(is):xseg(is)+lseg(is)-1) = .false.
          enddo
       endif

! Recompute medsmooth and badtest around that column

       call avMedFilter(map, goodPixel, width, medsmooth, badtest,  &
                        cbad(ic)-w2, cbad(ic)+w2, wy0, wy1)
    enddo
    deallocate(pixrate,goodcol,goodcur)

  end subroutine findBadCol

! Find bad rows and bad segments within
! Update medsmooth and badtest

  subroutine findBadRow(map, goodPixel, wid1D, maxproba, maxratio, minratio,  &
                medsmooth, badtest, width, xbad, ybad, yext, tbad, nbad)

! map       : Image array
! goodPixel : False for bad pixels
! wid1D     : Width over which the comparison is done on rows
! maxproba  : Probability threshold for false detection
! maxratio  : Relative detection threshold for dark features
! minratio  : Relative detection threshold for bright features
! medsmooth : Image after median smoothing
! badtest   : Significance map
! width     : Width over which the comparison is done on image
! xbad      : array of RAWX coordinates
! ybad      : array of RAWY coordinates
! yext      : array of RAWY extensions
! tbad      : array of bad pixel types
! nbad      : number of bad pixels

    integer(kind=int32), dimension(:,:), pointer       :: map
    logical,             dimension(:,:), intent(inout) :: goodPixel
    real,                dimension(:,:), intent(inout) :: medsmooth, badtest
    integer,                             intent(in)    :: wid1D, width
    real,                                intent(in)    :: maxproba, maxratio, &
                                                          minratio
    integer(kind=int16), dimension(:),   intent(inout) :: xbad,ybad,yext,tbad
    integer,                             intent(inout) :: nbad

    integer, dimension(:),   allocatable :: weight
    integer, dimension(1024)             :: cbad, ubad, xseg, lseg
    real,    dimension(:),   allocatable :: pixrate, profil, toCounts
    logical, dimension(:),   allocatable :: goodrow, goodcur
    integer :: nx, ny, nbc, nseg, nbd, ngood, w21D, w2
    integer :: ic, is, i, j, j1, jc
    real    :: refcts

    call message(AppMsg,Verbose,"Looking for bad rows...")

    w21D = wid1D / 2
    w2 = width / 2

! Must exchange X and Y window parameters

    i = wx0
    wx0 = wy0
    wy0 = i
    i = wx1
    wx1 = wy1
    wy1 = i

! Project the image onto RAWY and look for bad rows
! Compensate for brightness profile along rows

    nx = size(map,1)
    ny = size(map,2)
    allocate(goodcur(wid1D))
    allocate(profil(ny),toCounts(ny),goodrow(ny))
    allocate(pixrate(nx),weight(nx))

    toCounts = 0.
    profil = 0.
    goodrow = any(goodPixel,1)
    do j = wx0, wx1
       if (.not. goodrow(j))  cycle
       j1 = min(max(j,wx0+w21D),wx1-w21D)
       weight = count(goodPixel(:,j1-w21D:j1+w21D),2)
       ngood = count(weight > 0)
       where (weight == 0)
          weight = 1
       endwhere
       pixrate = real(sum(map(:,j1-w21D-1:j1+w21D-1),2,  &
                          goodPixel(:,j1-w21D:j1+w21D)))
       refcts = sum(pixrate/weight) * (wy1-wy0+1) / ngood
       if (refcts > 1.e-10) then
          toCounts(j) = sum(pixrate/weight,goodPixel(:,j)) / refcts
          if (toCounts(j) > 1.e-10)  &
             profil(j) = sum(map(:,j-1),goodPixel(:,j)) / toCounts(j)
       else
          toCounts(j) = 1.
       endif
    enddo

    call findBad1D(profil, toCounts, wid1D, maxproba, maxratio, minratio,  &
                   'row', cbad, ubad, nbc)

    nbd = count(ubad(1:nbc) == DEAD)
    write(errstr,*) nbd, " dark and", nbc-nbd, " bright rows found in image"
    call message(AppMsg,Sparse)
    deallocate(profil,toCounts,weight)

! Check whether only part of the row is bright

    pixrate = 0.
    do ic = 1, nbc
       write(errstr,*) "Row", cbad(ic)
       call message(AppMsg,Noisy)
       j1 = min(max(cbad(ic),wx0+w21D),wx1-w21D)

! Check number of nearby entirely bad rows (wx0 and wy0 are inverted)

       if (count(goodrow(j1-w21D:j1+w21D)) < 3) then
          write(errstr,*) "next to other bad rows. Declared entirely bad."
          call message(AppMsg,Noisy)
          do i = wy0, wy1
             nbad = nbad + 1
             xbad(nbad) = i
             ybad(nbad) = cbad(ic)
             yext(nbad) = 1
             tbad(nbad) = ubad(ic)
          enddo
          goodPixel(wy0:wy1,cbad(ic)) = .false.
       else

! Get expected count rate for each pixel (wx0 and wy0 are inverted)

          do i = wy0, wy1
             goodcur = goodPixel(i,j1-w21D:j1+w21D)

! Ignore nearby suspect rows starting at current one
! Previous ones should have been cleaned already

             do jc = ic, nbc
                if (abs(cbad(jc)-j1) <= w21D)  &
                   goodcur(w21D+1+cbad(jc)-j1) = .false.
             enddo
             pixrate(i) = real(sum(map(i-1,j1-w21D-1:j1+w21D-1),goodcur))  &
                        / max(count(goodcur),1)
          enddo

          call findBadSeg(map(:,cbad(ic)-1), goodPixel(:,cbad(ic)),  &
                          pixrate, ubad(ic), xseg, lseg, nseg)
          do is = 1, nseg
             do i = 0, lseg(is)-1
                nbad = nbad + 1
                xbad(nbad) = xseg(is) + i
                ybad(nbad) = cbad(ic)
                yext(nbad) = 1
                tbad(nbad) = ubad(ic)
             enddo
             goodPixel(xseg(is):xseg(is)+lseg(is)-1,cbad(ic)) = .false.
          enddo
       endif
    enddo
    deallocate(pixrate,goodrow,goodcur)

! Reset X and Y window parameters

    i = wx0
    wx0 = wy0
    wy0 = i
    i = wx1
    wx1 = wy1
    wy1 = i

! Recompute medsmooth and badtest around each bad row
! (must be done when wx0/wx1/wy0/wy1 have their usual meaning)

    do ic = 1, nbc
       call avMedFilter(map, goodPixel, width, medsmooth, badtest,  &
                        wx0, wx1, cbad(ic)-w2, cbad(ic)+w2)
    enddo

  end subroutine findBadRow

! Find statistical peaks in a 1-D distribution (both positive and negative)
! Compare with respect to neighbourhood

  subroutine findBad1D(profil, toCounts, wid1D, maxproba, maxratio, minratio, &
                       lab1D, cbad, tbad, nbc)

! profil   : Number of counts in each bin (corrected for missing pixels)
! toCounts : Coefficient to convert to counts (where Poisson statistics apply)
! wid1D    : Width over which the comparison is done
! maxproba : Probability threshold for false detection
! maxratio : Relative detection threshold for dark features
! minratio : Relative detection threshold for bright features
! lab1D    : Label (row, column) for messages
! cbad     : List of peaks identified
! tbad     : List of bad pixel types (HOT or DEAD)
! nbc      : Number of peaks identified

    real,    dimension(:),    intent(in)  :: profil, toCounts
    real,                     intent(in)  :: maxproba, maxratio, minratio
    integer,                  intent(in)  :: wid1D
    character(len=*),         intent(in)  :: lab1D
    integer, dimension(1024), intent(out) :: cbad, tbad
    integer,                  intent(out) :: nbc

    real,                  external    :: erfc
    real,    dimension(:), allocatable :: medsmooth, badtest
    logical, dimension(:), allocatable :: goodCol
    integer, dimension(1)              :: maxpos
    real    :: colmu, proba, proba2, mu2
    real    :: pixrate, minsigma
    integer :: i, i0, ibad, ncur, nx, w21D
    logical :: negative

    w21D = wid1D / 2

! Get number of sigmas corresponding to maxproba

    minsigma = (quantilePoisson(1.e4,maxproba)-1.e4) / 100

    nbc = 0

    if (wx1-wx0+1 < wid1D) then
       write(errstr,*) "Width for local ", lab1D, " analysis", wid1D,  &
                    " is larger than window size", wx1-wx0+1, ". Nothing done."
       call warning("findBad1D10")
       return
    endif

! No need to go further if all the pixels are already bad

    if (maxval(toCounts) < 1.e-10)  return

! Get the average number of counts per pixel

    nx = size(profil)
    allocate(medsmooth(nx), badtest(nx), goodCol(nx))
    goodCol = toCounts > 1.e-10

! Get the median and the test statistic over wid1D pixels. First iteration

    badtest = 0.
    do i = wx0, wx1
       medsmooth(i) = getMedian(profil, goodCol, wid1D, wx0, wx1, i,  &
                                .false., lab1D, ncur, cbad, tbad, nbc)
       if (ncur == 0)  cycle
       badtest(i) = setTest1D(profil, goodCol, toCounts(i), medsmooth(i),  &
                              i, wid1D)
!       if (i == 147) then
!          write(errstr,*) "profil, goodCol around", i, ":",  &
!                          profil(i-4:i+4), goodCol(i-4:i+4)
!          call message(AppMsg,Noisy)
!          write(errstr,*) "medsmooth, badtest at", i, ":",  &
!                           medsmooth(i), badtest(i)
!          call message(AppMsg,Noisy)
!       endif
    enddo

    do ibad = 1,nx

! Look for maximum local excess where goodCol is True
! Should not be used with pointers

       maxpos = maxloc(abs(badtest),goodCol)

! Positive or negative peak ?

       if (badtest(maxpos(1)) > minsigma) then
          negative = .false.
       else if (badtest(maxpos(1)) < (-minsigma)) then
          negative = .true.
       else
          write(errstr,*) "Peak at", maxpos, badtest(maxpos(1)),  &
                          "sigma <", minsigma
          call message(AppMsg,Noisy)
          exit   ! Stop if less than required significance
       endif

       goodCol(maxpos(1)) = .false.
       write(errstr,*) "Investigating", maxpos, badtest(maxpos(1))
       call message(AppMsg,Noisy)
       pixrate = medsmooth(maxpos(1))

       i0 = min(max(maxpos(1),wx0+w21D),wx1-w21D)
       ncur = count(goodCol(i0-w21D:i0+w21D))

! Multiply by coefficients to convert to counts with Poisson statistics
! erfc fails when its argument is larger than 9

       colmu = pixrate * toCounts(maxpos(1))
       if (abs(badtest(maxpos(1))) < 12.5) then
          proba = erfc(abs(badtest(maxpos(1)))/sqrt(2.)) / 2.
       else
          proba = 0.
       endif
       mu2 = profil(maxpos(1)) * toCounts(maxpos(1))

! Compute exact significance using binomial law

       if (negative) then
          proba2 = cumulativeBinomial(nint(ncur*colmu),nint(mu2),  &
                                      ncur/(ncur+1.d0))
       else
          proba2 = cumulativeBinomial(nint(mu2),nint(ncur*colmu),1.d0/(ncur+1))
       endif
       proba = max(proba,proba2)

       write(errstr,*) "Pixel value:", profil(maxpos(1)),  &
                       "   Local median:", colmu/toCounts(maxpos(1)),  &
                       "   Probability:", proba
       call message(AppMsg,Noisy)

       if (proba > maxproba) then

! Reinstate. Was not so bad after all.

          goodCol(maxpos(1)) = .true.
          badtest(maxpos(1)) = 0.
          cycle
       else

! Recompute local median (excluding current pixel) and significance
! Estimate local dispersion on sqrt(profil) excluding current pixel

          do i = max(maxpos(1)-w21D,wx0),min(maxpos(1)+w21D,wx1)
             if (goodCol(i)) then
                medsmooth(i) = getMedian(profil, goodCol, wid1D, wx0, wx1, i, &
                                        negative, lab1D, ncur, cbad, tbad, nbc)
                if (ncur == 0) then
                   badtest(i) = 0.
                   cycle
                endif
                badtest(i) = setTest1D(profil, goodCol, toCounts(i),  &
                                       medsmooth(i), i, wid1D)
             else
                badtest(i) = 0.
             endif
          enddo
       endif

! Count dark columns only if they are below maxratio * average
! Count bright columns only if they are above minratio * average

       if (negative .and. profil(maxpos(1))>maxratio*pixrate) then
          write(errstr,*) "Counts no less than", maxratio,  &
                          " times local average. Not rejected."
          call message(AppMsg,Noisy)
          cycle
       endif
       if (.not. negative .and. profil(maxpos(1))<minratio*pixrate) then
          write(errstr,*) "Counts no more than", minratio,  &
                          " times local average. Not rejected."
          call message(AppMsg,Noisy)
          cycle
       endif
       nbc = nbc + 1
       cbad(nbc) = maxpos(1)
       if (negative) then
          tbad(nbc) = DEAD
       else
          tbad(nbc) = HOT
       endif
    enddo

    deallocate(medsmooth, badtest, goodCol)

  end subroutine findBad1D

! Prepares search for bad rows/columns by computing test statistic
! around each pixel
! Estimate the dispersion of the local counts
! Systematically add 1 to variance to avoid false detections at low counts
! Estimate the local excesses in rms units
! to avoid detecting wrong bright columns next to grey ones

  real function setTest1D(profil, goodCol, toCounts, medloc, i, wid1D)

! profil   : Number of counts in each bin (corrected for missing pixels)
! goodCol  : set to False for rows/columns already set to bad
! toCounts : Coefficient to convert to counts (where Poisson statistics apply)
! medloc   : Local median
! i        : current pixel index
! wid1D    : width over which comparison is done

    real,    dimension(:),    intent(in)  :: profil
    logical, dimension(:),    intent(in)  :: goodCol
    integer,                  intent(in)  :: i, wid1D
    real,                     intent(in)  :: toCounts, medloc

    real,    dimension(:),    allocatable :: profloc
    logical, dimension(:),    allocatable :: goodcur
    integer, dimension(1)                 :: maxpos
    real    :: medloc2, varsqrt1, varsqrt2, vartot, testPoisson, nloc, mu2
    integer :: i0, w21D, ncur

! Return immediately if current pixel is already bad

    if (.not. goodCol(i)) then
       setTest1D = 0.
       return
    endif

    w21D = wid1D / 2
    i0 = min(max(i,wx0+w21D),wx1-w21D)
    ncur = count(goodCol(i0-w21D:i0+w21D)) - 1

! First call without broadening nor rejecting brightest pixel
! Similar to what was done for the median

!    allocate(goodcur(2*w21D+1),profloc(2*w21D+1))
!    goodcur = goodCol(i0-w21D:i0+w21D)
!    profloc = profil(i0-w21D:i0+w21D)
!    goodcur(w21D+1+i-i0) = .false.
!    if (medloc > 100) then
!       varsqrt1 = getWidth(sqrt(profloc),goodCol=goodcur)
!    else
!       varsqrt1 = getWidth(sqrt(pack(profloc,goodcur)))
!    endif
!    deallocate(goodcur,profloc)
    varsqrt1 = 1e3   ! Algorithm above looks too permissive

! Increase window if too few valid points for variance

!    if (ncur < max(wid1D-2,5)) then
    if (ncur < wid1D-1) then
       w21D = w21D + 1
       i0 = min(max(i,wx0+w21D),wx1-w21D)
       ncur = count(goodCol(i0-w21D:i0+w21D)) - 1
    endif

    allocate(goodcur(2*w21D+1),profloc(2*w21D+1))
    goodcur = goodCol(i0-w21D:i0+w21D)
    profloc = profil(i0-w21D:i0+w21D)

! Always ignore central pixel

    goodcur(w21D+1+i-i0) = .false.

! Also ignore brightest pixel for variance calculation (coupled bright pixels)

    maxpos = maxloc(profloc,goodcur)
    goodcur(maxpos(1)) = .false.
    ncur = ncur - 1

! Use local dispersion as security for Gaussian test
! Remove gradient only if enough counts (> 100)

    if (medloc > 100) then
       varsqrt2 = getWidth(sqrt(profloc),goodCol=goodcur)
    else
       varsqrt2 = getWidth(sqrt(pack(profloc,goodcur)))
    endif
!    if (i == 147) then
!       write(errstr,*) "sqrt(profloc), goodcur:", sqrt(profloc), goodcur
!       call message(AppMsg,Noisy)
!       write(errstr,*) "varsqrt at i =", i, ":", varsqrt2
!       call message(AppMsg,Noisy)
!    endif       

! Take the smallest dispersion, but variance should be at least Poisson

    vartot = max(2*min(varsqrt1,varsqrt2),1.)**2
    medloc2 = medloc * toCounts
    nloc = profil(i) * toCounts
    setTest1D = (nloc - medloc2) / sqrt(medloc2*vartot+1)

    deallocate(goodcur,profloc)

! Poisson test for positive excesses
! Li&Ma 1983, ApJ 272, 317 Eq(17)

    if (setTest1D > 3) then
       mu2 = (nloc + ncur*medloc2) / (ncur+1)
       testPoisson = sqrt(2*(nloc*log(nloc/mu2) +  &
                             ncur*medloc2*log(max(medloc2,1./ncur)/mu2)))
       setTest1D = min(setTest1D,testPoisson)
    endif

  end function setTest1D

! Find bright segments within a bright column/row
! The Kolmogorov-Smirnov approach used to flag warm areas based on source counts
! should work better than this

  subroutine findBadSeg(profil, goodPixel, pixrate, btype, xseg, lseg, nseg)

! profil    : Number of counts in each pixel
! goodPixel : False if pixel is already known to be bad
! pixrate   : Expected count rate per pixel along column/row
! btype     : Type of bad column/row (HOT or DEAD)
! xseg      : List of first pixels in segments
! lseg      : List of segment lengths
! nseg      : Number of segments identified

    integer, dimension(:),    intent(in)  :: profil
    logical, dimension(:),    intent(in)  :: goodPixel
    real,    dimension(:),    intent(in)  :: pixrate
    integer,                  intent(in)  :: btype
    integer, dimension(1024), intent(out) :: xseg, lseg
    integer,                  intent(out) :: nseg

    real,    dimension(:),    allocatable :: expcounts, boxcounts
    logical, dimension(:),    allocatable :: goodcur
    integer, dimension(1)                 :: maxpos

! Confidence level for stopping accumulating bad segments

    real, parameter  :: conflevel = 0.01

    integer :: wseg, ws2, ws20, nctmax, nx, nctmin, nleft, nloc, cleft, zero
    integer :: i, i0, i1, j0, j1, is
    real    :: exploc
    logical :: negative

    negative = btype == DEAD

! Prepare integration with boxcar window

    nx = size(profil)
    allocate(goodcur(nx),boxcounts(nx),expcounts(nx))

! Define minimum width to have at least 1 expected count
! wseg should not be larger than the window size

    if (negative) then
       cleft = max(sum(profil,goodPixel),1)
       nleft = count(goodPixel)
       wseg = int(nleft/real(cleft)) + 1
       ws20 = max(min(wseg,wy1-wy0-1)/2,1)
    else
       ws20 = 0
    endif

! Ignore bad pixels to handle case of entirely bad rows

    ws2loop: do ws2 = ws20,(wy1-wy0-1)/2
       wseg = 2 * ws2 + 1
       do i = wy0+ws2,wy1-ws2
          nleft = count(goodPixel(i-ws2:i+ws2))
          expcounts(i) = sum(pixrate(i-ws2:i+ws2),goodPixel(i-ws2:i+ws2))
          if (expcounts(i)*wseg < nleft)  cycle ws2loop
          expcounts(i) = max(expcounts(i),1.)
       enddo
       exit
    enddo ws2loop

    write(errstr,*) "Minimum segment length:", wseg
    call message(AppMsg,Noisy)

! Define value at which pixels will never be found bad

    if (negative) then
       zero = cleft*wseg + 1
    else
       zero = 0
    endif

    goodcur = goodPixel
    boxcounts = zero
    do i = wy0+ws2,wy1-ws2
       if (any(goodPixel(i-ws2:i+ws2))) then
          nloc = sum(profil(i-ws2:i+ws2),goodPixel(i-ws2:i+ws2))
          boxcounts(i) = (nloc - expcounts(i)) / sqrt(expcounts(i))
       endif
    enddo

! Pick segments starting with maximum value

    nseg = 0
    do is = 1, nx

       nleft = count(goodcur)
       if (nleft == 0)  exit

! Find position of maximum and maximum number of counts acceptable

       if (negative) then
          maxpos = minloc(boxcounts)
          if (boxcounts(maxpos(1)) == zero)  exit
          nctmax = quantilePoisson(sum(pixrate,goodcur),-conflevel)
       else
          maxpos = maxloc(boxcounts)
          if (boxcounts(maxpos(1)) == zero)  exit
          nctmax = quantilePoisson(sum(pixrate,goodcur),conflevel)
       endif
       i0 = maxpos(1) - ws2
       i1 = maxpos(1) + ws2

! If nctmax = 0 then reject whole column

       cleft = sum(profil,goodcur)
       if ((nctmax==0 .and. .not.negative) .or. (cleft==0 .and. negative)) then
          nseg = 1
          xseg(nseg) = 1
          lseg(nseg) = nx
          goodcur = .false.
          exit
       endif

! Test whether remaining number of counts is reasonable
! Exit if local count rate is reasonable as well or no counts left

       if (negative) then
          if (cleft >= nctmax) then
             nctmin = quantilePoisson(sum(pixrate(i0:i1),goodcur(i0:i1)),  &
                                      -wseg*conflevel/max(nleft,wseg))
             nloc = sum(profil(i0:i1),goodcur(i0:i1))
             write(errstr,*) "Counts in segment:", nloc, " at", maxpos(1), &
                             "   Minimum:", nctmin
             call message(AppMsg,Noisy)
             if (nloc >= nctmin) then
                if (nloc == 0 .and. nseg > 0) then

! Exit only if not adjacent to rejected area
! While nloc=0 the segments accumulate to the right

                   if (i0 > xseg(nseg)+lseg(nseg))  exit
                else
                   exit
                endif
             endif
          endif
       else if (cleft <= nctmax) then
          if (cleft == 0) then
             exit
          else
             nctmin = quantilePoisson(sum(pixrate(i0:i1),goodcur(i0:i1)),  &
                                      conflevel/nleft)
             nloc = sum(profil(i0:i1),goodcur(i0:i1))
             write(errstr,*) "Counts in segment:", nloc, " at", maxpos(1), &
                             "   Maximum:", nctmin
             call message(AppMsg,Noisy)
             if (nloc <= nctmin)  exit
          endif
       endif

! Extend to near border or other segment if less than wseg good pixels
! Some bad pixels may be isolated (found as individual bad pixels)
! Segments may overlap

       if (i0 >= wy0+wseg) then
          if (.not. all(goodcur(i0-wseg:i0-1))) then
             do i = i0-1,i0-wseg,-1
                if (.not. goodcur(i))  exit
             enddo
             i0 = i + 1
          endif
       endif
       if (i0 < wy0+wseg)  i0 = wy0
       if (i1 <= wy1-wseg) then
          if (.not. all(goodcur(i1+1:i1+wseg))) then
             do i = i1+1,i1+wseg
                if (.not. goodcur(i))  exit
             enddo
             i1 = i - 1
          endif
       endif
       if (i1 > wy1-wseg)  i1 = wy1

       if (all(goodcur(i0:i1))) then
          nseg = nseg + 1
          xseg(nseg) = i0
          lseg(nseg) = i1 - i0 + 1
       else

! Respect type of previous bad pixels

          j1 = i0 - 2
          do while (j1 < i1-1)

! Find first formerly good pixel

             do j0 = j1+2,i1
                if (goodcur(j0))  exit
             enddo
             if (j0 > i1)  exit

! Find first already declared bad pixel

             do j1 = j0+1,i1
                if (.not. goodcur(j1))  exit
             enddo
             j1 = j1 - 1
             nseg = nseg + 1
             xseg(nseg) = j0
             lseg(nseg) = j1 - j0 + 1
          enddo
       endif

! Set new segment to bad

       goodcur(i0:i1) = .false.

! Recompute boxcounts, so that next segment is found as far as possible

       boxcounts(i0:i1) = zero
       do j0 = max(i0-ws2-1,wy0+ws2),i0-1
          if ( goodcur(j0) )  exit
       enddo
       if (j0 < i0) then
          boxcounts(j0+1:i0-1) = zero
          boxcounts(j0) = sum(profil(j0-ws2:j0+ws2),goodcur(j0-ws2:j0+ws2))
          exploc = expcounts(j0) * count(goodPixel(j0-ws2:j0+ws2)) / wseg
          boxcounts(j0) = (boxcounts(j0) - exploc) / sqrt(exploc)
       endif
       do j1 = min(i1+ws2+1,wy1-ws2),i1+1,-1
          if ( goodcur(j1) )  exit
       enddo
       if (j1 > i1) then
          boxcounts(i1+1:j1-1) = zero
          boxcounts(j1) = sum(profil(j1-ws2:j1+ws2),goodcur(j1-ws2:j1+ws2))
          exploc = expcounts(j1) * count(goodPixel(j1-ws2:j1+ws2)) / wseg
          boxcounts(j1) = (boxcounts(j1) - exploc) / sqrt(exploc)
       endif
    enddo

    if (nleft == 0) then
       write(errstr,*) "Found entirely bad."
    else
       write(errstr,*) "Identifying segments of width", wseg, " down to", &
                       cleft, " counts and", nleft, " pixels left."
    endif
    call message(AppMsg,Noisy)

    deallocate(goodcur,boxcounts,expcounts)

  end subroutine findBadSeg

! Apply median/average filter on a fraction of a map

  subroutine avMedFilter(map, goodPixel, width, medsmooth, badtest,  &
                         imin, imax, jmin, jmax)

! map       : Image array
! goodPixel : False for bad pixels (inout because temporarily set to false)
! width     : Width over which the comparison is done
! medsmooth : Image after median/average smoothing
! badtest   : Map of estimated significance
! imin,imap : Bounds on abscissa (truncated to window anyway)
! jmin,jmax : Bounds on ordinate (truncated to window anyway)

    integer(kind=int32), dimension(:,:), pointer       :: map
    logical,             dimension(:,:), intent(inout) :: goodPixel
    integer,                             intent(in)    :: width, imin, imax,  &
                                                          jmin, jmax
    real,                dimension(:,:), intent(inout) :: medsmooth, badtest

    integer, dimension(:), allocatable :: packSpi
    real    :: signif
    integer :: ncur, w2
    integer :: i, j, i1, i2, j1, j2

    w2 = width / 2
    allocate(packSpi(width**2))

    do j = max(jmin,wy0),min(jmax,wy1)
       j1 = max(j-w2,wy0)
       j2 = min(j+w2,wy1)
       do i = max(imin,wx0),min(imax,wx1)
          i1 = max(i-w2,wx0)
          i2 = min(i+w2,wx1)
          if (goodPixel(i,j)) then
             ncur = count(goodPixel(i1:i2,j1:j2))
             if (ncur < width) then
                badtest(i,j) = 100.   ! Will be removed in findBadPix
             else

! Compute local median, not counting central pixel
! map starts at 0
! pack does not work well without a third argument

                goodPixel(i,j) = .false.
                medsmooth(i,j) = avMedian(pack(map(i1-1:i2-1,j1-1:j2-1),  &
                                             goodPixel(i1:i2,j1:j2),packSpi), &
                                          map(i-1,j-1), signif, ncur-1)
                badtest(i,j) = signif
                goodPixel(i,j) = .true.
             endif
          endif
       enddo
    enddo
    deallocate(packSpi)

  end subroutine avMedFilter

! Apply local average or median filter on dataref
! Return median when data0 is dark
! Return smaller of median + 1 or average when data0 is bright
! Also return significance of data0, using Li&Ma '83

  real function avMedian(dataref, data0, signif, nval)

! dataref   : reference data array
! data0     : data point to test against dataref
! signif    : significance (in sigma units) of data point
! ncur      : number of data points if dataref is larger

    integer(kind=int32), dimension(:), intent(in)  :: dataref
    integer,                           intent(in)  :: data0
    real,                              intent(out) :: signif
    integer, optional,                 intent(in)  :: nval

    real    :: mu, mu2, varsqrt, sigPoisson
    integer :: ncur

    if (present(nval)) then
       ncur = nval
    else
       ncur = size(dataref)
    endif

! Get median local value

    mu = localMedian(dataref(1:ncur))

! Get observed dispersion
! sqrt(median(x)) is a good approximation to median(sqrt(x))
! Never correct for gradient in 2D

    varsqrt = 2 * getWidth(sqrt(real(dataref(1:ncur))), medCol=sqrt(mu))
    varsqrt = max(varsqrt,.1)   ! Avoid division by 0

    if (data0 < mu) then

! Always use gaussian limit in the case of a dark pixel

       signif = (data0-mu) / (sqrt(mu)*varsqrt)
    else

! Get average local value outside suspect pixel
! (better than median when very small)

       mu2 = real(sum(dataref(1:ncur))) / ncur
       mu = min(mu+1,mu2)

! Apply both Gauss and Poisson tests, whichever gives lower significance
! Li&Ma 1983, ApJ 272, 317 Eq(17)

       signif = (data0-mu) / (sqrt(mu+1./ncur)*varsqrt)
       if (signif > 3) then
          mu2 = (data0 + ncur*mu) / (ncur+1)
          sigPoisson = sqrt(2*(data0*log(data0/mu2) +  &
                               ncur*mu*log(max(mu,1./ncur)/mu2)))
          signif = min(signif,sigPoisson)
       endif
    endif
    avMedian = mu

  end function avMedian

! Get local median in 1-D array, excluding central pixel
! If not negative, return smaller of median+1 and average
! Declare pixel as bad if no neighbour left

  real function getMedian(rateCol, goodCol, length, wx0, wx1, i, negative,  &
                          lab1D, ncur, cbad, tbad, nbc)

! rateCol  : input array
! goodCol  : list of indices to consider
! length   : length for median averaging
! wx0,wx1  : Bounds within which the median should be taken
! i        : current index
! negative : Look for dead pixels instead if true
! lab1D    : Label (row, column) for messages
! ncur     : number of good pixels in window (not counting central pixel)
! cbad     : List of peaks identified
! tbad     : List of bad pixel types (HOT or DEAD)
! nbc      : Number of peaks identified

    real,    dimension(:),    intent(in)    :: rateCol
    logical, dimension(:),    intent(inout) :: goodCol
    integer,                  intent(in)    :: length, wx0, wx1, i
    logical,                  intent(in)    :: negative
    character(len=*),         intent(in)    :: lab1D
    integer,                  intent(out)   :: ncur
    integer, dimension(1024), intent(inout) :: cbad, tbad
    integer,                  intent(inout) :: nbc

    real,    dimension(:),    allocatable   :: rate
    logical, dimension(:),    allocatable   :: goodcur
    real    :: aver
    integer :: i0, len2

! Compute number of good pixels in vicinity

    i0 = min(max(i-length/2,wx0),wx1-length+1)
    ncur = count(goodCol(i0:i0+length-1))
    if (goodCol(i))  ncur = ncur - 1

! Declare bad and return if no neighbour left
! Don't do it for central call (length = 3)

    if (length > 3 .and. goodCol(i) .and. ncur==0) then
       getMedian = 0.
       write(errstr,*) "No ", lab1D, " left around", i, ". Counted as bad."
       call message(AppMsg,Sparse)
       nbc = nbc + 1
       cbad(nbc) = i
       if (negative) then
          tbad(nbc) = DEAD
       else
          tbad(nbc) = HOT
       endif
       goodCol(i) = .false.
       return
    endif

! Extend range if too few pixels and not at border already
! Don't do it for central call (length = 3)

    len2 = length
    if (length > 3 .and. ncur <= 2 .and. i0 > wx0 .and. i0+length <= wx1) then
       len2 = len2 + 2
       i0 = i0 - 1
    endif

    allocate(rate(len2),goodcur(len2))
    rate = rateCol(i0:i0+len2-1)
    goodcur = goodCol(i0:i0+len2-1)

! Ignore central pixel

    goodcur(i-i0+1) = .false.
    ncur = count(goodcur)

! Remove underlying gradient if > 100

    if (ncur < 2) then
       getMedian = sum(rate,goodcur)
    else
       getMedian = localMedian(pack(rate,goodcur))
       if (getMedian > 100) then
          call subGrad(rate,goodcur,i-i0-(len2-1)/2)
          getMedian = localMedian(pack(rate,goodcur))
       endif
    endif

! If not negative, return average, if between median and median+1

    if (.not.negative) then
       aver = sum(rate,goodcur) / ncur
       getMedian = max(getMedian,min(getMedian+1,aver))
    endif

    deallocate(rate,goodcur)

  end function getMedian

! Get width of the distribution of a 1-D array
! Use mean absolute deviation (more robust than variance against outliers)
! Account for a possible gradient if the optional argument goodCol is passed
! Abscissae are assumed to be equidistant for the gradient
! Compute median internally if the optional argument medCol is not passed

  real function getWidth(rateCol, goodCol, medCol)

! rateCol  : input array
! goodCol  : list of indices to consider
! medCol   : median of rateCol(goodCol)

    real,    dimension(:), intent(in)           :: rateCol
    logical, dimension(:), intent(in), optional :: goodCol
    real,                  intent(in), optional :: medCol

    real,    dimension(:),    allocatable   :: rate
    real    :: aver, width1, width2
    integer :: length, ncur, ndof
    logical :: withgrad, withmed

! Are optional arguments present ?

    withgrad = present(goodCol)
    withmed = present(medCol)

! Number of good pixels and of degrees of freedom

    length = size(rateCol)
    if (withgrad) then
       ncur = count(goodCol)
       ndof = ncur - 1
    else
       ndof = length
    endif

! Return 0 if not enough points

    if (ndof <= 1) then
       getWidth = 0.
       return
    endif

    if (withgrad) then

! First estimate without subtracting gradient

       if (withmed) then
          aver = medCol
       else
          aver = localMedian(pack(rateCol,goodCol))
       endif
       width1 = sum(abs(rateCol-aver),mask=goodCol) / ncur

! Remove average and underlying gradient

       allocate(rate(length))
       rate = rateCol
       call subGrad(rate,goodCol)
       aver = localMedian(pack(rate,goodCol))
       width2 = sum(abs(rate-aver),mask=goodCol) / ndof
!       if (.not. any(goodCol(1:5))) then
!          write(errstr,*) "rate-aver, goodCol:", rate-aver, goodCol
!          call message(AppMsg,Noisy)
!          write(errstr,*) "width1, width2:", width1, width2
!          call message(AppMsg,Noisy)
!       endif
       deallocate(rate)

! Take smaller of both estimates

       getWidth = min(width1,width2)
    else

! Straight estimate of mean absolute deviation

       if (withmed) then
          aver = medCol
       else
          aver = localMedian(rateCol)
       endif
       getWidth = sum(abs(rateCol-aver)) / ndof
    endif

! Mean absolute deviation gives 0.8 sigma for Gauss or Poisson distributions

    getWidth = getWidth / 0.8

  end function getWidth

! Remove underlying gradient from an array

  subroutine subGrad(rateCol, goodCol, offset)

! rateCol  : input/output array
! goodCol  : list of indices to consider
! offset   : index offset with respect to (1+size(rateCol))/2 (default is 0)

    real,    dimension(:),    intent(inout) :: rateCol
    logical, dimension(:),    intent(in)    :: goodCol
    integer, optional,        intent(in)    :: offset

    real,    dimension(:),    allocatable   :: dist
    integer, dimension(1)                   :: posmin, posmax
    real    :: aver, grad, iav, i0, imin, imax
    integer :: k, pmin, pmax, length, ncur
    logical :: withgrad

! Check coherence of entries

    length = size(rateCol)
    if (size(goodCol) /= length) then
       write(errstr,*) "The two arguments do not have the same size",  &
                       length, size(goodCol), ". Contact developer."
       call error("subGrad01")
    endif

! Compute number of good pixels in vicinity

    ncur = count(goodCol)

! Remove underlying gradient if at least 3 points
! Do it only if first and last pixels are extrema, or if the distribution
! of points is symmetric (no gradient correction on median)

    if (ncur > 2) then
       allocate(dist(length))
       dist = (/ (real(k), k=1,length) /)
       imin = minval(dist,goodCol)
       imax = maxval(dist,goodCol)

! Check optional argument

       if (present(offset)) then
          i0 = (1+length)/2 + offset
       else
          i0 = (imin+imax) / 2
       endif
       withgrad = abs(sum(dist-i0,goodCol)) < 0.1
       if (.not. withgrad) then
          posmin = minloc(rateCol,goodCol)
          posmax = maxloc(rateCol,goodCol)
          pmin = min(posmin(1),posmax(1))
          pmax = max(posmin(1),posmax(1))
          withgrad = max(abs(pmin-imin),abs(pmax-imax)) < 0.1
       endif
       if (withgrad) then
          dist = dist - i0
          aver = sum(rateCol,goodCol) / ncur
          iav = sum(dist,goodCol) / ncur
          grad = sum((rateCol-aver)*(dist-iav),goodCol)  &
               / sum((dist-iav)**2,goodCol)
          rateCol = rateCol - grad * dist
       endif
       deallocate(dist)
    endif

  end subroutine subGrad

! Get the uploaded bright pixels and the dead pixels from the CCF
! Also get the bad columns switched off by large offsets
! goodPixel should be all True where map > 0 on input

  subroutine getCalBadpix(map, withdead, withbrightccf, goodPixel,  &
                          xbad,ybad,yext,tbad,fbad,nbad)

! map       : Image array
! goodPixel : array of good pixels
! withdead : Dead pixels are written to output list if true
! withbrightccf: Bright pixels in CCF are ignored in the analysis
! xcal     : array of RAWX coordinates
! ycal     : array of RAWY coordinates
! yext     : array of RAWY extensions
! tbad     : array of bad pixel types
! fbad     : array of bad pixel status (uplinked, CCF or new)
! nbad     : number of bad pixels

    integer(kind=int32), dimension(:,:), pointer       :: map
    logical, intent(in)    :: withdead, withbrightccf
    logical,             dimension(:,:), intent(inout) :: goodPixel
    integer(kind=int16), dimension(:),   intent(inout) :: xbad, ybad, yext,  &
                                                          tbad, fbad
    integer, intent(inout) :: nbad
    integer :: nbad2 ! temp array to print out few messages

    integer(kind=int16), dimension(:), pointer :: rawX, rawY, ctype, yextent
    integer(kind=int16), dimension(:), pointer :: caloffX, caloffY
    logical,             dimension(:), pointer :: uplinked
    integer(kind=int16), dimension(:), allocatable :: xcal, ycal, yextcal,  &
                                                      tcal, fcal
    integer :: ibad, xcur, ycur, ymax, yhot
    integer :: locsum, xloc, yloc, ncal, oldbad
    integer :: underx, overx, undery, overy
    integer :: x_size_window, y_size_window
    integer :: windowx0, windowy0, windowdx, windowdy
    type(ModeParameterDataType) :: modeParameters

    call CAL_getModeParameters(modeParameters)
    underx = modeParameters%underscanX
    overx = modeParameters%overscanX
    undery = modeParameters%underscanY
    overy = modeParameters%overscanY
    windowx0 =  modeParameters%windowX0
    windowy0 =  modeParameters%windowY0
    windowdx =  modeParameters%windowDX
    windowdy =  modeParameters%windowDY
    x_size_window = windowdx + underx + overx
    y_size_window = windowdy + undery + overy
    write(errstr,*) "underx,overx,undery,overy,x_size_window,y_size_window:",&
                     underx,overx,undery,overy,x_size_window,y_size_window
    call message(AppMsg, Noisy)

! Read bad pixels declared in CCF

    call CAL_getBadPixelList(rawX, rawY, ctype, yextent, uplinked)
    ncal = size(rawX)
    write(errstr,*) ncal, "bad pixel entries in CCF"
    call message(AppMsg, Noisy)

! Transform uplinked into an integer array for BADFLAG
! Use intermediate array because mergeBad may increase its size

    ibad = 2*ncal + windowdx
    allocate(xcal(ibad), ycal(ibad), yextcal(ibad), tcal(ibad), fcal(ibad))
    xcal(1:ncal) = rawX
    ycal(1:ncal) = rawY
    yextcal(1:ncal) = yextent
    tcal(1:ncal) = ctype
    where (uplinked)
       fcal(1:ncal) = 1
    elsewhere
       fcal(1:ncal) = 2
    endwhere
    call CAL_releaseMemory(rawX, rawY)
    call CAL_releaseMemory(ctype, yextent)
    call CAL_releaseMemory(uplinked)

! Read offsets

    nullify(caloffY, caloffX)
    call CAL_getMOSoffsets(caloffY, caloffX)

! Detect large offsets

    if (size(caloffX) /= x_size_window .or. size(caloffY) /= y_size_window) then
       write(errstr,*) "The size of the array returned by CAL_getMOSoffsets",  &
            size(caloffX), size(caloffY), "is not what was expected",  &
            x_size_window, y_size_window, &
            ". Contact developer. Offsets are ignored."
       call warning("getCalBadpix11")
    else

! Consider only patched columns
! Patched rows would require adding windowdx entries for each

       do xcur = underx, underx+windowdx-1
          if (caloffX(xcur) > 4000) then
             ncal = ncal + 1
             xcal(ncal) = xcur - underx + windowx0
             ycal(ncal) = windowy0
             yextcal(ncal) = windowdy
             tcal(ncal) = HOT
             fcal(ncal) = 1
             write(errstr,*) "Column", xcal(ncal), "declared bad via offsets"
             call message(AppMsg, Noisy)
          endif
       enddo
    endif
    call CAL_releaseMemory(caloffX,caloffY)

! Order and group the CAL bad pixels

    call mergeBad(xcal, ycal, yextcal, tcal, fcal, ncal)

    oldbad = nbad
    do ibad = 1, ncal

! Use standard integers for indices

       xcur = xcal(ibad)
       ycur = ycal(ibad)

! Ignore bad pixels outside current window
! Set them not uplinked to avoid failures in following loop

       if (ycur < wy0) then
          yextcal(ibad) = yextcal(ibad) - wy0 + ycur
          ycur = wy0
       endif
       if (xcur<wx0 .or. xcur>wx1 .or. ycur>wy1 .or. yextcal(ibad)<=0) then
          fcal(ibad) = 2
          cycle
       endif
       yextcal(ibad) = min(yextcal(ibad),int(wy1-ycur+1,kind=int16))

! Indicate uplinked bad pixels in the 2-D array

       if (fcal(ibad)==1 .or. tcal(ibad)==DEAD .or. withbrightccf)  &
          goodPixel(xcur,ycur:ycur+yextcal(ibad)-1) = .false.

! Optionally write dead pixels to list

       if (withdead .and. tcal(ibad)==DEAD) then
          nbad = nbad + 1
          xbad(nbad) = xcur
          ybad(nbad) = ycur
          yext(nbad) = yextcal(ibad)
          tbad(nbad) = tcal(ibad)
          fbad(nbad) = fcal(ibad)
       endif
    enddo

! Check uplinked bad pixels are absent from the data (only check maximum)
! 2nd loop allows to know whether neighbouring pixels are declared bad as well

    do ibad = 1, ncal
       if (fcal(ibad) == 1) then
          xcur = xcal(ibad)
          ycur = ycal(ibad)
          ymax = ycur
          do yhot = ycur+1, ycur+yextcal(ibad)-1
             if (map(xcur-1,ymax-1) < map(xcur-1,yhot-1))  ymax = yhot
          enddo

! Clear error only when maximum > sum(neighbours) because of pattern projection
! Diagonal events may contribute as well

          locsum = 0
          do yloc = max(ymax-1,wy0), min(ymax+1,wy1)
             do xloc = max(xcur-1,wx0), min(xcur+1,wx1)
                if (goodPixel(xloc,yloc))  locsum = locsum + map(xloc-1,yloc-1)
             enddo
          enddo

          if (map(xcur-1,ymax-1) > locsum) then
             write(errstr,*) "Bright pixel at", xcur, ymax,  &
                " is declared uplinked in the CCF but is present in the data."
             call warning("getCalBadpix10")

! Reactivate search for the whole column segment

             goodPixel(xcur,ycur:ycur+yextcal(ibad)-1) = .true.
          endif
       endif
    enddo
    deallocate(xcal, ycal, yextcal, tcal, fcal)

    nbad2 = min(oldbad+20,nbad)

    if (withdead .and. nbad2 > oldbad) then
       write(errstr,*) "Included", nbad-oldbad, " dead pixels in list. Printing first 20..."
       call message(AppMsg, Verbose)
       write(errstr,*) "Xdead:", xbad(oldbad+1:nbad2)
       call message(AppMsg, Noisy)
       write(errstr,*) "Ydead:", ybad(oldbad+1:nbad2)
       call message(AppMsg, Noisy)
       write(errstr,*) "Yext:", yext(oldbad+1:nbad2)
       call message(AppMsg, Noisy)
    endif

  end subroutine getCalBadpix

! Parameter reading

  subroutine getParamValues(ev_map, bad_tab, incremental, maxproba,  &
       maxratio, minratio, findcols, findbright, finddead, w2, w21D,  &
       niter, withcal, withdead, withbrightccf)

! ev_map      : Handle to the input image
! bad_tab     : Handle to the output bad pixels table
! incremental : Add new bad pixels to bad pixels file if true
! maxproba    : Probability threshold for false detection
! maxratio    : Relative detection threshold for dark features
! minratio    : Relative detection threshold for bright features
! findcols    : Look for bad rows or columns if true
! findbright  : Look for bright pixels if true
! finddead    : Look for dead pixels if true
! w2          : Half width of window for 2D search
! w21D        : Half width of window for 1D search
! niter       : Number of iterations
! withcal     : CAL_getBadPixelList is called if true
! withdead    : Dead pixels are written to output list if true
! withbrightccf: Bright pixels in CCF are ignored in the analysis

    type(ArrayT),        intent(out) :: ev_map
    type(TableT),        intent(out) :: bad_tab
    real,                intent(out) :: maxproba, maxratio, minratio
    logical,             intent(out) :: withcal, withdead, findcols,  &
                                        findbright, finddead, incremental,  &
                                        withbrightccf
    integer,             intent(out) :: w2, w21D, niter

    type(DataSetT) :: im_set, bad_set
    integer, dimension(:), pointer :: sizarray
    character(len=1024) :: image_out, bad_out
    integer :: node
    character(len=8) :: tabname

! Open input image file

    image_out = stringParameter(outimage)
    write(errstr,*) "Opening image file : ",trim(image_out)
    call message(AppMsg,Sparse)
    im_set = dataSet(image_out, Read)
    ev_map = array(im_set, "PRIMARY")
    sizarray => dimensions(ev_map)

! Get window parameters

    if (hasAttribute(ev_map,"WINDOWX0")) then
       wx0 = int32Attribute(ev_map,"WINDOWX0")
       wx1 = wx0 + int32Attribute(ev_map,"WINDOWDX") - 1
       wy0 = int32Attribute(ev_map,"WINDOWY0")
       wy1 = wy0 + int32Attribute(ev_map,"WINDOWDY") - 1
       if (wx0 > 1) then
          write(errstr,*) "Window limits:", wx0, wx1, wy0, wy1
          call message(AppMsg,Sparse)
       endif
       wx0 = max(1,wx0)
       wx1 = min(sizarray(0),wx1)
       wy0 = max(1,wy0)
       wy1 = min(sizarray(1),wy1)
    else
       wx0 = 1
       wx1 = sizarray(0)
       wy0 = 1
       wy1 = sizarray(1)
    endif

! Width

    w2 = intParameter(hwidth2D)
    write(errstr,*) "Half width for 2D searches:", w2
    call message(AppMsg,Sparse)

! Threshold

    maxproba = realParameter(threshold)
    write(errstr,*) "False detection probability:",maxproba," per pixel"
    call message(AppMsg,Sparse)

! Columns

    findcols = booleanParameter(badcols)
    if (findcols) then
       call message(AppMsg,Sparse,  &
          "The task will also look for bad segments in rows or columns")
       w21D = intParameter(hwidth1D)
       write(errstr,*) "Half width for 1D searches:", w21D
       call message(AppMsg,Sparse)
    endif

! Bright pixels

    minratio = 1.
    findbright = booleanParameter(onbright)
    if (findbright) then
       call message(AppMsg,Sparse,"The task will look for bright pixels")
       minratio = realParameter(minrat)
       write(errstr,*) "The task will look for bright pixels more than",  &
                       minratio, " relative to their neighbours"
       call message(AppMsg,Sparse)
    endif

! Dead pixels

    maxratio = 1.
    finddead = booleanParameter(ondead)
    if (finddead) then
       call message(AppMsg,Sparse,"The task will look for dark pixels")
       maxratio = realParameter(maxrat)
       write(errstr,*) "The task will look for dark pixels less than",  &
                       maxratio, " relative to their neighbours"
       call message(AppMsg,Sparse)
    endif

! Output file name

    bad_out   = stringParameter(outbadpix)
    incremental = booleanParameter(incbadpix)

    if (incremental) then
       bad_set = dataSet(bad_out, Modify)
       bad_tab = table(bad_set, "BADPIX")

! Check compatibility of keywords between map and bad pixels file

       if (.not. equalKeywords(attributable(ev_map),attributable(bad_tab), &
                               keycompat,.false.) .or.  &
           .not. equalKeywords(attributable(im_set),attributable(bad_set), &
                               keycompat,.false.)) then
          write(errstr,*) "The bad pixels file ", trim(bad_out),  &
             " does not correspond to the image ", trim(image_out),  &
             ". Some keywords differ among ",keycompat
          call error("getParamValues03")
       endif
       write(errstr,*) "Bad pixels will be read from and written to ",  &
                       trim(bad_out)
       call message(AppMsg,Sparse)
    else
       write(errstr,*) "Bad pixels will be written to ",trim(bad_out)
       call message(AppMsg,Sparse)
       bad_set = dataSet(bad_out, Create)
       call setAttributes(attributable(bad_set),attributable(im_set))
       if ( (trim(stringAttribute(ev_map,"INSTRUME")) .eq. 'RGS1') .or. &
        (trim( stringAttribute(ev_map,"INSTRUME") ) .eq. 'RGS2') ) then
            node = 0
            if (hasAttribute(ev_map,"CCDNODE")) then
                node = int32Attribute(ev_map,"CCDNODE")
            endif
            tabname = "BADPIX0"
            if ( node == 1) then
                tabname = "BADPIX1"
            endif
       else
            tabname = "BADPIX"
       endif

       bad_tab = addTable(bad_set, tabname, 0)
       call setAttributes(attributable(bad_tab), attributable(ev_map))
    endif

! Number of iterations

    niter = intParameter(iterate)

! Set CAL state

    withcal = booleanParameter(usecal)

    if (withcal) then
       call CAL_setState(parent(ev_map))
       withdead = booleanParameter(adddead)
       if (withdead)  call message(AppMsg,Sparse,  &
          "Dead pixels from the CAL will be included in the bad pixels list")
       withbrightccf = .not. booleanParameter(nobrightccf)
       if (.not. withbrightccf)  call message(AppMsg,Sparse,  &
          "Bright pixels are not read from the CAL")
    endif

  end subroutine getParamValues

end module embadpixfind_mod
